@@ -7,13 +7,13 @@
 """ Callbacks rules classes and helpers
 """
 import logging
-from datetime import datetime
+from functools import wraps
 
-from ._vendors.wrapt import FunctionWrapper
 from .condition_evaluator import ConditionEvaluator, is_condition_empty
 from .constants import LIFECYCLE_METHODS
 from .remote_exception import raw_traceback_formatter, traceback_formatter
 from .runtime_storage import runtime
+from .utils import now
 
 LOGGER = logging.getLogger(__name__)
 
@@ -49,58 +49,59 @@ class BaseCallback(object):
         return self.INTERRUPTIBLE
 
 
-def check_condition_wrapper(rule_data, function, condition, lifecycle):
+def check_condition_wrapper(wrapped, callback, condition, lifecycle):
     """ Wrapper that will check lifecycle method pre-condition before
     calling it.
     If pre-conditions are true, call the lifecycle method, otherwise
     return None.
     """
 
-    def wrapper(wrapped, callback, args, kwargs):
+    @wraps(wrapped)
+    def wrapper(inst, args, kwargs, **options):
         """ Wrapper around lifecycle method
         """
-
-        # Get the storage from the callback if available
-        storage = callback.storage if callback is not None else runtime
+        storage = callback.storage
 
         # Compute return value depending on the lifecycle method we wrap
-        return_value = kwargs.get("result") if lifecycle == "post" \
-            else kwargs.get("exc_info")
+        rv = options.get("result" if lifecycle == "post" else "exc_info")
 
         binding_eval_args = {
             "request": storage.get_current_request(),
             "response": storage.get_current_response(),
-            "inst": args[0],
-            "args": storage.get_current_args(args[1]),
-            "kwargs": args[2],
-            "data": rule_data,
-            "rv": return_value,
+            "inst": inst,
+            "args": storage.get_current_args(args),
+            "kwargs": kwargs,
+            "data": callback.data,
+            "rv": rv,
         }
 
         # Check the pre condition
         condition_result = condition.evaluate(**binding_eval_args)
 
-        LOGGER.debug(
-            "ConditionWrapper (%s) result for %s: %s",
-            condition,
-            function,
-            condition_result,
-        )
-
         if condition_result in (False, None):
+            LOGGER.debug(
+                "Not running %r on %r: %r is %s",
+                lifecycle,
+                callback,
+                condition,
+                condition_result,
+            )
             return None
 
+        LOGGER.debug("Running %r on callback %r", lifecycle, callback)
         # Execute the hook otherwise with the original args
-        return wrapped(*args, **kwargs)
+        return wrapped(inst, args, kwargs, **options)
 
-    return FunctionWrapper(function, wrapper)
+    wrapper.__wrapped__ = wrapped
+    return wrapper
 
 
-def call_count_wrapper(function, lifecycle, observation_key):
+def call_count_wrapper(wrapped, callback, lifecycle, observation_key):
     """ Wrapper around lifecycle methods that record number of calls
     """
 
-    def wrapper(wrapped, callback, args, kwargs):
+    @wraps(wrapped)
+    def wrapper(*args, **kwargs):
         """ Record the number of calls for this callback lifecycle method.
         Buffer the number in the callback itself (self.call_counts) and record
         an observation every X times, X being the field call_count_interval of
@@ -118,10 +119,11 @@ def call_count_wrapper(function, lifecycle, observation_key):
 
         return wrapped(*args, **kwargs)
 
-    return FunctionWrapper(function, wrapper)
+    wrapper.__wrapped__ = wrapped
+    return wrapper
 
 
-def perf_monitoring_wrapper(function, name):
+def perf_monitoring_wrapper(wrapped, storage, name):
     """ Wrapper that will monitor the performance of the rule callback.
     It will time the execution of the callback and send it as an
     observation metric.
@@ -129,16 +131,15 @@ def perf_monitoring_wrapper(function, name):
     The wrapper should be called first before every others.
     """
 
-    def wrapper(wrapped, callback, args, kwargs):
+    @wraps(wrapped)
+    def wrapper(*args, **kwargs):
         """ Time the callback execution.
         """
-        # Get the storage from the callback if available
-        storage = callback.storage if callback is not None else runtime
-
         with storage.trace(name):
             return wrapped(*args, **kwargs)
 
-    return FunctionWrapper(function, wrapper)
+    wrapper.__wrapped__ = wrapped
+    return wrapper
 
 
 class RuleCallback(BaseCallback):
@@ -151,9 +152,10 @@ class RuleCallback(BaseCallback):
         hook_name,
         rule_name,
         rulespack_id,
-        block,
-        test,
         runner,
+        beta=False,
+        block=False,
+        test=False,
         strategy_cls=None,
         data=None,
         conditions=None,
@@ -162,8 +164,9 @@ class RuleCallback(BaseCallback):
         metrics=None,
         payload_sections=None,
         priority=None,
+        attack_render_info=None,
+        attack_type=None,
         storage=runtime,
-        attack_render_info=None
     ):
         super(RuleCallback, self).__init__(
             hook_module, hook_name, strategy_cls
@@ -174,6 +177,7 @@ class RuleCallback(BaseCallback):
         if data is None:
             data = {}
         self.data = data
+        self.beta = beta
         self.block = block
         self.test = test
         self.runner = runner
@@ -185,9 +189,10 @@ class RuleCallback(BaseCallback):
         self.conditions = conditions
 
         if callbacks is None:
-            callbacks = []
+            callbacks = {}
         self.callbacks = callbacks
         self.attack_render_info = attack_render_info
+        self.attack_type = attack_type
 
         # Callbacks
         self.call_count_interval = call_count_interval
@@ -216,8 +221,9 @@ class RuleCallback(BaseCallback):
             rulespack_id=rule_dict["rulespack_id"],
             strategy_cls=rule_dict["hookpoint"].get("strategy", "import_hook"),
             attack_render_info=rule_dict["hookpoint"].get("attack_render_info", {}),
-            block=rule_dict["block"],
-            test=rule_dict["test"],
+            beta=rule_dict.get("beta", False),
+            block=rule_dict.get("block", False),
+            test=rule_dict.get("test", False),
             data=rule_dict.get("data"),
             conditions=rule_dict.get("conditions"),
             callbacks=rule_dict.get("callbacks"),
@@ -226,6 +232,7 @@ class RuleCallback(BaseCallback):
             metrics=rule_dict.get("metrics", []),
             payload_sections=rule_dict.get("payload"),
             priority=rule_dict.get("priority"),
+            attack_type=rule_dict.get("attack_type"),
             storage=storage,
         )
 
@@ -257,14 +264,13 @@ class RuleCallback(BaseCallback):
         """
         for lifecycle in LIFECYCLE_METHODS.values():
             conditions = self.conditions.get(lifecycle)
-
-            if not is_condition_empty(conditions) and hasattr(self, lifecycle):
+            lifecycle_method = getattr(self, lifecycle, None)
+            if lifecycle_method is not None and \
+                    not is_condition_empty(conditions):
                 conditions = ConditionEvaluator(conditions)
-
                 # Wrap the lifecycle method
-                lifecycle_method = getattr(self, lifecycle)
                 wrapped = check_condition_wrapper(
-                    self.data, lifecycle_method, conditions, lifecycle)
+                    lifecycle_method, self, conditions, lifecycle)
                 setattr(self, lifecycle, wrapped)
 
     def _apply_call_count(self):
@@ -273,8 +279,8 @@ class RuleCallback(BaseCallback):
             return
 
         for lifecycle in LIFECYCLE_METHODS.values():
-            if hasattr(self, lifecycle):
-                lifecycle_method = getattr(self, lifecycle)
+            lifecycle_method = getattr(self, lifecycle, None)
+            if lifecycle_method is not None:
 
                 observation_key = "%s/%s/%s" % (
                     self.rulespack_id,
@@ -283,7 +289,7 @@ class RuleCallback(BaseCallback):
                 )
 
                 wrapped = call_count_wrapper(
-                    lifecycle_method, lifecycle, observation_key)
+                    lifecycle_method, self, lifecycle, observation_key)
                 setattr(self, lifecycle, wrapped)
 
     def _apply_perf_monitoring(self):
@@ -292,8 +298,8 @@ class RuleCallback(BaseCallback):
             return
 
         for lifecycle in LIFECYCLE_METHODS.values():
-            if hasattr(self, lifecycle):
-                lifecycle_method = getattr(self, lifecycle)
+            lifecycle_method = getattr(self, lifecycle, None)
+            if lifecycle_method is not None:
 
                 metric_name = "sq.{}.{}".format(self.rule_name, lifecycle)
 
@@ -308,7 +314,7 @@ class RuleCallback(BaseCallback):
                     )
 
                 wrapped = perf_monitoring_wrapper(
-                    lifecycle_method, metric_name)
+                    lifecycle_method, self.storage, metric_name)
                 setattr(self, lifecycle, wrapped)
 
     def exception_infos(self, infos={}):
@@ -319,12 +325,15 @@ class RuleCallback(BaseCallback):
     def record_attack(self, infos=None, at=None):
         """Record an attack."""
         if at is None:
-            at = datetime.utcnow()
+            at = now()
         payload = {
             "infos": infos,
             "rulespack_id": self.rulespack_id,
             "rule_name": self.rule_name,
+            "beta": self.beta,
+            "block": self.block,
             "test": self.test,
+            "attack_type": self.attack_type,
             "time": at,
         }
         if "context" in self.payload_sections:
@@ -336,25 +345,18 @@ class RuleCallback(BaseCallback):
 
     def record_observation(self, metric_name, key, value, at=None):
         """Record a metric observation."""
-
-        self.global_record_observation(metric_name, key, value, at, storage=self.storage)
-
-    @classmethod
-    def global_record_observation(cls, metric_name, key, value, at=None, storage=runtime):
-        """Record a metric observation"""
-
         if at is None:
-            at = datetime.utcnow()
+            at = now()
         payload = (metric_name, at, key, value)
         LOGGER.debug("Observed metric %r", payload)
-        storage.observe("observations", payload, report=False)
+        self.storage.observe("observations", payload, report=False)
 
     def record_exception(self, exc, exc_info, stack=None, infos=None, at=None):
         """Record an exception."""
         if infos is None:
             infos = {}
         if at is None:
-            at = datetime.utcnow()
+            at = now()
         # Try to recover some infos from the exception if it's a
         # SqreenException.
         try:
@@ -369,6 +371,7 @@ class RuleCallback(BaseCallback):
             "klass": exc_info[0].__name__,
             "infos": infos,
             "rule_name": self.rule_name,
+            "rulespack_id": self.rulespack_id,
             "test": self.test,
             "time": at,
             "backtrace": bt,
@@ -376,16 +379,10 @@ class RuleCallback(BaseCallback):
         LOGGER.debug("Observed exception %r", payload)
         self.storage.observe("sqreen_exceptions", payload, self.payload_sections)
 
-    def record_overtime(self, lifecycle=None, at=None):
+    def record_overtime(self, lifecycle, at=None):
         """Record an overtime execution."""
-        if at is None:
-            at = datetime.utcnow()
-        name = self.rule_name
-        if lifecycle is not None:
-            name += ".{}".format(lifecycle)
-        payload = ("request_overbudget_cb", at, name, 1)
-        LOGGER.debug("Observed overbudget callback %r", payload)
-        self.storage.observe("overtime_callbacks", payload, report=False)
+        self.storage.record_overtime(
+            "{}.{}".format(self.rule_name, lifecycle), at=at)
 
     @property
     def lifecycle_methods(self):
@@ -403,4 +400,6 @@ class RuleCallback(BaseCallback):
         return self.storage.get_whitelist_match() or False
 
     def __repr__(self):
-        return "%s(rule_name=%r)" % (self.__class__.__name__, self.rule_name)
+        return "%s(rule_name=%r, hook_module=%r, hook_name=%r, strategy=%r)" % (
+            self.__class__.__name__, self.rule_name, self.hook_module,
+            self.hook_name, self.strategy)
